/* 
 *       Scientific Applications & Visualization Group
 *            SAVG web page: http://math.nist.gov/mcsd/savg/
 *       Mathematical & Computational Science Division
 *            MCSD web page: http://math.nist.gov/mcsd/
 *       National Institute of Standards and Technology
 *            NIST web page: http://www.nist.gov/
 * 
 *  
 *  This software was developed at the National Institute of Standards
 *  and Technology by employees of the Federal Government in the course
 *  of their official duties. Pursuant to title 17 Section 105 of the
 *  United States Code this software is not subject to copyright
 *  protection and is in the public domain.  isolate is an experimental 
 *  system.  NIST assumes no responsibility whatsoever for its use by 
 *  other parties, and makes no guarantees, expressed or implied, about 
 *  its quality, reliability, or any other characteristic.
 *  
 *  We would appreciate acknowledgement if the software is used.
 */

/*
 * pflgt.C: $Revision: 1.1 $ $Date: 2003/4/02
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include <Performer/pf.h>
#include <Performer/pfutil.h>
#include <Performer/pfdu.h>
#include <Performer/pf/pfScene.h>
#include <Performer/pf/pfGroup.h>
#include <Performer/pr/pfMaterial.h>
#include <Performer/pr.h>
#include <Performer/pr/pfGeoSet.h>
#include <Performer/pf/pfDCS.h>
#include <Performer/pf/pfLightSource.h>
#include "parse.C"
#include <fstream.h>
#include "getString.C"

/*
 * pfdLoadFile_lgt -- Load ".lgt" data files into IRIS Performer
 */
#define DIFFUSE_R (1.0f)
#define DIFFUSE_G (1.0f)
#define DIFFUSE_B (0.8f)

#define AMBIENT_R (0.4f)
#define AMBIENT_G (0.4f)
#define AMBIENT_B (0.3f)

extern "C" pfNode *
pfdLoadFile_lgt (char *fileName)
{
    FILE	*lgtFile	= NULL;
    double	 ra		= 1.0;
    double	 ga		= 1.0;
    double	 ba		= 1.0;
    double	 rd		= 1.0;
    double	 gd		= 1.0;
    double	 bd		= 1.0;
    char	 *nextWord;
    char         *lite_type;
    char        **groupnames = NULL;
    float        intensity      = 1.0;
    pfVec3	 lite_loc;
    float        w              = 0.0;
    pfVec3	 lite_dir;
    float        dropOffRate    = 0.0;
    float        cutOffAngle    = 0.0;
    int          branches       = 0;
    int          putbranch      = 0;
    int          maxLightSources = 9;
    char *line = new char[2047];
    pfLightSource* lsource;
    pfGroup **group = NULL;
    pfGroup *topgroup = new pfGroup;
    group = (pfGroup **) realloc(group, sizeof(pfGroup*)*maxLightSources);
    groupnames = (char **) realloc(groupnames, sizeof(char*)*maxLightSources);
    /* restore builder to initial state */
    pfdResetBldrGeometry();
    pfdResetBldrState();

    /* open ".lgt" file */
    FILE *fp= pfdOpenFile(fileName) ;
    if (!fp)
      return NULL;
    lite_type = new char[3];
    /* loop through the file */
    int pargc ;
    char **pargv ;
    while ((getString(&line,fp)!=EOF)) {
      dpf_charsToArgcArgv(line, &pargc, &pargv) ;
      // if this is a comment line - continue
      if (pargc < 1) continue;
      nextWord = pargv[0];
      if (strlen(nextWord) < 3) continue;

      // reset default values
      intensity = 1.0;
      lite_loc[PF_X] = 0.0;
      lite_loc[PF_Y] = 0.0;
      lite_loc[PF_Z] = 0.0;
      lite_dir[PF_X] = 0.0;
      lite_dir[PF_Y] = 1.0;
      lite_dir[PF_Z] = 0.0;
      w = 0.0;
      ra = AMBIENT_R;
      ga = AMBIENT_G;
      ba = AMBIENT_B;
      rd = DIFFUSE_R;
      gd = DIFFUSE_G;
      bd = DIFFUSE_B;
      dropOffRate= 0.0;
      cutOffAngle = 40.0;
      putbranch = -1;
      // check for one of the light types
      if (nextWord[0] == 'D' || nextWord[0] == 'S' || nextWord[0] == 'P') {

	// assign a type
	lite_type[0] = nextWord[0];
	// read to the next comment line
	int c = getString(&line,fp);
	if (c == EOF)
	  break;
	dpf_charsToArgcArgv(line, &pargc, &pargv) ;
	while(pargc > 0 && pargv[0][0] != '#') {
	  nextWord = pargv[0];
	  // read intensity
	  if (strcasecmp(nextWord,"intensity")==0) {
	    intensity = atof(pargv[1]);
	  }
	  
	  // read color
	  if (strcasecmp(nextWord,"color")==0) {
	    if (pargc == 4) {
	      ra = atof(pargv[1]);
	      ga = atof(pargv[2]);
	      ba = atof(pargv[3]);
	      rd = ra;
	      gd = ga;
	      bd = ba;
	    }
	    else if (pargc > 4) {
	      if (strcasecmp(pargv[4],"diffuse")==0) {
		rd = atof(pargv[1]);
		gd = atof(pargv[2]);
		bd = atof(pargv[3]);
	      }
	      else if (strcasecmp(pargv[4],"ambient")==0) {
		ra = atof(pargv[1]);
		ga = atof(pargv[2]);
		ba = atof(pargv[3]);
	      }
	      else {
		ra = atof(pargv[1]);
		ga = atof(pargv[2]);
		ba = atof(pargv[3]);
		rd = ra;
		gd = ga;
		bd = ba;
	      }
	    }
	  }

	  // direction
	  if (strcasecmp(nextWord,"direction")==0) {
	    lite_dir[PF_X] = atof(pargv[1]);
	    lite_dir[PF_Y] = atof(pargv[2]);
	    lite_dir[PF_Z] = atof(pargv[3]);
	  }

	  // location
	  if (strcasecmp(nextWord,"location")==0) {
	    lite_loc[PF_X] = atof(pargv[1]);
	    lite_loc[PF_Y] = atof(pargv[2]);
	    lite_loc[PF_Z] = atof(pargv[3]);
	  }

	  // dropOffRate
	  if (strcasecmp(nextWord,"dropOffRate")==0) {
	    dropOffRate = atof(pargv[1]);
	  }

	  // cutOffAngle
	  if (strcasecmp(nextWord,"cutOffAngle")==0) {
	    cutOffAngle = atof(pargv[1]);
	  }

	  if (strcasecmp(nextWord,"group")==0) {
	    // this section defines the group node name
	    // if this is loaded into DIVERSE code, 
	    // the names world,ether,nav,scene have defined meanings

	    // first see if this is the only branch
	    if (branches == 0) {
	      groupnames[0] = pargv[1];
	      group[0] = new pfGroup;
	      group[0]->setName(pargv[1]);
	      putbranch = 0;
	      ++branches;
	    }
	    else {
	      // see if this group is already in use
	      int found = -1;
	      for ( int j = 0; j < branches; ++j ) {
		if (pargv[1][0] == groupnames[j][0]) found = j;
	      }
	      if (found == -1) {
		// set up a new group 
		groupnames[branches] = pargv[1];
		group[branches] = new pfGroup;
		group[branches]->setName(pargv[1]);
		putbranch = branches;
		++branches;
	      }
	      else {
		// put light in an old branch
		putbranch = found;
	      }
	    }
	  }

	  // read next line
	  int c = getString(&line,fp);
	  if (line!=NULL && c != EOF) {
	    dpf_charsToArgcArgv(line, &pargc, &pargv) ;
	  }
	  else {
	    pargc = 0;
	  }
	} // end while

	// create the light source
	lsource = new pfLightSource();
	// set intensity throught the color
	ra *= intensity;
	ga *= intensity;
	ba *= intensity;
	rd *= intensity;
	gd *= intensity;
	bd *= intensity;
	lsource->setColor(PFLT_AMBIENT, ra,ga,ba);
	lsource->setColor(PFLT_DIFFUSE, rd,gd,bd);

	// assign a w value for directional and spotlights
	if (lite_type[0] == 'P') w = 1.0;
	if (lite_type[0] == 'S') w = 1.0;
	lsource->setPos(lite_loc[PF_X],lite_loc[PF_Y],lite_loc[PF_Z],w);

	// set direction if applicable
	if (lite_type[0] == 'S')
	  lsource->setSpotDir(lite_dir[PF_X],lite_dir[PF_Y],lite_dir[PF_Z]);

	if (lite_type[0] == 'S')
	  lsource->setSpotCone(dropOffRate, cutOffAngle);

	// add light to branch
	if ( putbranch == -1 ) {
	  // no named branch

	  // see if there is already a unnamed branch
	  int found = -1;
	  for ( int j = 0; j < branches; ++j ) {
	    if (groupnames[j] == NULL) found = j;
	  }
	  if (found == -1) {
	    // set up a new group 
	    groupnames[branches] = NULL;
	    group[branches] = new pfGroup;
	    putbranch = branches;
	    ++branches;
	  }
	}
	group[putbranch]->addChild(lsource);

      }
      line = NULL;
}

    fclose(fp);
    /* print statistics */
    pfNotify(PFNFY_NOTICE, PFNFY_PRINT, "pfdLoadFile_lgt: %s", fileName);
    pfNotify(PFNFY_NOTICE, PFNFY_PRINT, "  Enhanced (by SGS) 7/11/2002 6:30pm");
    // put the branches on the top node and return
    for ( int j = 0; j < branches; ++j ) {
      topgroup->addChild(group[j]);
    }
    return (pfGroup*)topgroup;
}
